/*

  HelloX.C
  ========
  (c) Copyright Paul Griffiths 1999
  Email: paulgriffiths@cwcom.net

  "Hello, World!", X Window System style.
  
  Hacked into a Graphite demo helloworld

*/


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <iconv.h>

/*########################################################*/
/*# BEGIN:  TT/AA message */
#include <X11/Xft/Xft.h>
/*# END : TT/AA message */
/*########################################################*/
#include "DrawGraphiteString.h"

/* Types, structs and unions */
union UniBuffer {
	char      utf8[128];
	FcChar16  utf16[64];
};


/*  Global variables  */
namespace {

Display     * display;
int           screen_num;
char        * appname;
	
void report_failure(const std::string msg) {
	std::cerr << appname << ": " << msg << std::endl;
	exit(EXIT_FAILURE);
}

size_t utf16length(const FcChar16 *str) {
	const FcChar16 *i = str;
	while (*i != 0) ++i;
	return i - str;
}

void      display_freetype_message(XftDraw *, XftFont *, FcChar16 *, int msg_x, int msg_y);
void      convert_utf8_to_utf16(const std::string from, FcChar16 * to);
XftFont * open_font(const char *font_name, unsigned int height);

}

int main( int argc, char * argv[] ) 
{
	/*  Window variables  */
	Window       win;
	int          x, y;
	unsigned int width, height;
	unsigned int border_width;
	char *       window_name = "Hello, X Window System!";
	char *       icon_name   = "HelloX";


	/*  Display variables  */
	char *       display_name = NULL;
	unsigned int display_width, display_height;


	/*  Miscellaneous X variables  */
	XSizeHints *  size_hints;
	XWMHints   *  wm_hints;
	XClassHint *  class_hints;
	XTextProperty windowName, iconName;
	XEvent        report;
	XFontStruct * font_info;
	XGCValues     values;
	GC            gc;

	/* Xft stuff */
	XftDraw *xftdraw;
	XftFont *xftfont;
	XftFont *xftfont2;

	appname = argv[0];


	FcChar16 msg[64] =  {
		0x1019, 0x1004, 0x103A, 0x1039, 0x1002, 0x101C, 
		0x102C, 0x1015, 0x102C, 0x1000, 0x1019, 0x1039, 
		0x1018, 0x102C, 0x104B, 0x0000
	};

	if (argc > 1 && strlen(argv[1]) < 100)
	{
		convert_utf8_to_utf16(argv[1], msg);
	}
	else
	{
		std::cout << "Using default message / font" << std::endl
		          << "Usage: " << argv[0] << " 'text' 'grfont.ttf'" << std::endl;
	}

	/*  Allocate memory for our structures  */
	if ( !( size_hints  = XAllocSizeHints() ) || 
	     !( wm_hints    = XAllocWMHints()   ) ||
	     !( class_hints = XAllocClassHint() ) ) {
		report_failure("couldn't allocate memory.");
	}


	/*  Connect to X server  */

	if ( (display = XOpenDisplay(display_name)) == NULL ) {
		report_failure(std::string("couldn't connect to X server") + display_name);
	}


	/*  Get screen size from display structure macro  */

	screen_num     = DefaultScreen(display);
	display_width  = DisplayWidth(display, screen_num);
	display_height = DisplayHeight(display, screen_num);


	/*  Set initial window size and position, and create it  */

	x = y = 0;
	width  = display_width / 3;
	height = display_width / 3;

	win = XCreateSimpleWindow(display, RootWindow(display, screen_num),
	x, y, width, height, border_width,
	BlackPixel(display, screen_num),
	WhitePixel(display, screen_num));


	/*  Set hints for window manager before mapping window  */

	if ( XStringListToTextProperty(&window_name, 1, &windowName) == 0 ) {
		report_failure("structure allocation for windowName failed");
	}

	if ( XStringListToTextProperty(&icon_name, 1, &iconName) == 0 ) {
		report_failure("structure allocation for iconName failed");
	}

	size_hints->flags       = PPosition | PSize | PMinSize;
	size_hints->min_width   = 200;
	size_hints->min_height  = 100;

	wm_hints->flags         = StateHint | InputHint;
	wm_hints->initial_state = NormalState;
	wm_hints->input         = True;

	class_hints->res_name   = appname;
	class_hints->res_class  = "hellox";

	XSetWMProperties(display, win, &windowName, &iconName, argv, argc,
	size_hints, wm_hints, class_hints);


	/*  Choose which events we want to handle  */
	XSelectInput(display, win, ExposureMask | KeyPressMask |
	ButtonPressMask | StructureNotifyMask);

	/*  Display Window  */
	XMapWindow(display, win);


	/* Setup xft fonts and drawbles. */
	const char * testFontName = "Padauk";

	if (argc > 2)
	{
		testFontName = argv[2]; // assumes ascii name
	}

	xftfont = open_font(testFontName, 25);	                      
	if (!IsGraphiteFont(xftfont)) {
		report_failure("XftFontOpen1(): graphite font not found"); 
	}

	xftfont2 = open_font("Std SILDoulos", 18);

	/*  Load a font called "9x15"  */
	if ( (font_info = XLoadQueryFont(display, "9x15")) == NULL ) {
		report_failure("cannot open 9x15 font.");
	}


	xftdraw = XftDrawCreate(display, (Drawable) win,
	                        DefaultVisual(display, screen_num),
	                        DefaultColormap(display, screen_num));
	if (!xftdraw) {
		report_failure("XftDrawCreate()");
	}

	/*  Create graphics context  */
	gc = XCreateGC(display, win, 0, &values);

	XSetFont(display, gc, font_info->fid);
	XSetForeground(display, gc, BlackPixel(display, screen_num));


	/*  Enter event loop  */
	while ( true ) {
		static char * message = "Hello, X Window System!";
		static int    length;
		static int    font_height;
		static int    msg_x, msg_y;

		XNextEvent(display, &report);

		switch ( report.type ) {
		case Expose:
			if ( report.xexpose.count != 0 )
				break;

			/*  Output message centrally in window  */
			length = XTextWidth(font_info, message, strlen(message));
			msg_x  = (width - length) / 2;

			font_height = font_info->ascent + font_info->descent;
			msg_y  = (height + font_height) / 2;

			XDrawString(display, win, gc, msg_x, msg_y,
			message, strlen(message));

			display_freetype_message(xftdraw, xftfont, msg, msg_x, msg_y + 30);

			break;


		case ConfigureNotify:
			/*  Store new window width & height  */
			width  = report.xconfigure.width;
			height = report.xconfigure.height;

			break;


		case ButtonPress:            /*  Fall through  */
		case KeyPress:
			/*  Clean up and exit  */
			XftFontClose (display, xftfont);
			XftFontClose (display, xftfont2);
			XftDrawDestroy(xftdraw);
			XFree(size_hints);
			XFree(wm_hints);
			XFree(class_hints);
			XUnloadFont(display, font_info->fid);
			XFreeGC(display, gc);
			XCloseDisplay(display);
			exit(EXIT_SUCCESS);
		}
	}

	return EXIT_SUCCESS;   /*  We shouldn't get here  */
}

namespace {

void display_freetype_message(XftDraw *xftdraw, XftFont *font, 
		FcChar16 * msg, int msg_x, int msg_y)
{
	XftColor color_fg;
	XColor dummyc,fg;
	Status retval;
//	int nglyph;
//	int nchar;
//	FT_UInt glyph[300];
//	int i;

	retval = XAllocNamedColor(display,DefaultColormap(display,screen_num),
	                          "black", &fg, &dummyc);
	if (!retval) {
		report_failure("XAllocNamedColor(): Failed"); 
	}

	color_fg.color.red = dummyc.red;
	color_fg.color.green = dummyc.green;
	color_fg.color.blue = dummyc.blue;
	color_fg.color.alpha = 0x00ff00;
	color_fg.pixel = fg.pixel;
	std::cout << std::hex << color_fg.color.red << ' '
	          << std::hex << color_fg.color.green << ' '
	          << std::hex << color_fg.color.blue << ' '
	          << std::hex << color_fg.color.alpha << std::endl;

	/* convert ucs4 to index
	replace with silgraphite code
	*/

//	char msg2[16] = "Hello World!!!\000";
	//DrawGraphiteString(xftdraw, xftfont, msg_x-80, msg_y+30, msg);
	//DrawGraphiteString8(xftdraw, xftfont2, msg_x+10, msg_y+70, msg2, 18);
	DrawGraphiteString(display, xftdraw, font, msg_x, msg_y, msg, utf16length(msg));
	/*
	for(i=0;i<nchar;i++)
	glyph[i] = XftCharIndex(display, xftfont, msg[i]);
	nglyph = nchar; */ /* for simple 1:1 char to glyph mapping */

	//	      XftDrawGlyphs(xftdraw, &color_fg, xftfont, msg_x-80, msg_y+30,
	//			      &glyph[0], nglyph);
	//	      XftDrawGlyphSpec(xftdraw, &color_fg, xftfont,
	//			      &glyph[0], nglyph);
}

void convert_utf8_to_utf16(const std::string from, FcChar16 * to)
{
	iconv_t ic = iconv_open("UTF-16","UTF-8");

	if (ic == iconv_t(-1))
	{
		std::cerr << "iconv open error" << std::endl;
	}
	
	size_t maxlenIn = from.length();
	
	char * in = const_cast<char*>(from.c_str());
	char * out = reinterpret_cast<char *>(to);
	size_t left = maxlenIn*2+1;

	std::cout << from << " char " << maxlenIn << std::endl;
	size_t cLen = iconv(ic,&in,&maxlenIn,&out,&left);
	std::cout << "found " << from << " " << cLen << " char " << from.length() << std::endl;
	iconv_close(ic);

	if (left > from.length())
	{
		report_failure("could not convert UTF-8 to UTF-16");
	}
}


XftFont *open_font(const char *font_name, unsigned int height)
{
	XftFont * xft_font = XftFontOpen(display, screen_num, 
			XFT_FAMILY, XftTypeString, font_name, 
			XFT_SIZE, XftTypeInteger, height, 
			NULL);
	if (!xft_font)
		report_failure(std::string("XftFontOpen(): Attempt to open \"") + font_name + "\" failed."); 
	return xft_font;
}

}
